from datetime import date, datetime

from mongoengine import Q

from calculator.tests.helpers import insert_prices_record_with_method, get_prices_with_int_date_record
from config import configure_app
from cryptomodel.coinmarket import prices, Coin, EUR, EURQuote, Status
from cryptomodel.fixer import exchange_rates
from cryptomodel.cryptostore import user_transaction
from calculator.BalanceCalculator import BalanceCalculator
from cryptomodel.helpers import do_connect

DATE_FORMAT = '%Y-%m-%d'


def test_one_transaction():
    the_prices = ({"XTZ": EUR(price=1)})

    xtz_trans = user_transaction(user_id=1, volume=31.182, symbol="XTZ", value=176.08408,
                                 price=2.44, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    bc = BalanceCalculator(transactions=[xtz_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")

    assert (out.converted_value == 31.182)


def test_no_symbols_should_not_throw_exception_should_mark_trans_invalid():
    the_prices = {}

    xtz_trans = user_transaction(user_id=1, volume=31.182, symbol="XTZ", value=176.08408,
                                 price=2.44, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    bc = BalanceCalculator(transactions=[xtz_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")
    assert (out.transactions[0].is_valid == False)
    assert (out.converted_value == 0)


def test_none_symbols_should_not_throw_exception_should_mark_trans_invalid():
    the_prices = None

    xtz_trans = user_transaction(user_id=1, volume=31.182, symbol="XTZ", value=176.08408,
                                 price=2.44, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    bc = BalanceCalculator(transactions=[xtz_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")
    assert (out.transactions[0].is_valid == False)
    assert (out.converted_value == 0)


def test_five_distinct_symbols():
    the_prices = {
        "XTZ": EUR(price=2.45),
        "ADA": EUR(price=.11862),
        "TRX": EUR(price=0.0161217),
        "XLM": EUR(price=0.08713060),
        "XRP": EUR(price=0.24919500),
    }

    xtz_trans = user_transaction(user_id=1, volume=31.182, symbol="XTZ", value=76.64,
                                 price=2.44, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")
    ada_trans = user_transaction(user_id=1, volume=2076.384678, symbol="ADA", value=244.968,
                                 price=.118, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")
    xlm_trans = user_transaction(user_id=1, volume=1796, symbol="XLM", value=149.068,
                                 price=0.08738980, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    xrp_trans = user_transaction(user_id=1, volume=1010.8, symbol="XRP", value=212.268,
                                 price=.21, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    trx_trans = user_transaction(user_id=1, volume=6420, symbol="TRX", value=103.5174,
                                 price=0.01612420, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    bc = BalanceCalculator(transactions=[xtz_trans, xrp_trans, ada_trans, xlm_trans, trx_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")

    assert (out.converted_value == 2.45 * 31.182
            + .11862 * 2076.384678
            + 0.0161217 * 6420
            + 0.08713060 * 1796
            + 0.24919500 * 1010.8)


def test_roi():
    the_prices = {
        "XTZ": EUR(price=2),
        "ADA": EUR(price=3),

    }

    xtz_trans = user_transaction(user_id=1, volume=30, symbol="XTZ", value=31.182,
                                 price=1, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")
    ada_trans = user_transaction(user_id=1, volume=20, symbol="ADA", value=40,
                                 price=2, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")

    bc = BalanceCalculator(transactions=[xtz_trans, ada_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")
    assert (len(out.user_grouped_symbol_values) == 2)
    assert out.user_grouped_symbol_values['ADA'].roi == ((3 * 20) - (40)) / (40)
    assert out.roi == ((2 * 30 + 3 * 20) - (40 + 31.182)) / (40 + 31.182)


def test_volume():
    the_prices = {
        "XTZ": EUR(price=2),
        "ADA": EUR(price=3),

    }

    xtz_trans = user_transaction(user_id=1, volume=30, symbol="XTZ", value=31.182,
                                 price=1, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="BUY")
    ada_trans = user_transaction(user_id=1, volume=20, symbol="XTZ", value=40,
                                 price=2, date="2020-07-15",
                                 source="kraken",
                                 source_id="",
                                 type="TRADE",
                                 order_type="SELL")

    bc = BalanceCalculator(transactions=[xtz_trans, ada_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")
    assert out.user_grouped_symbol_values['XTZ'].volume == 10


def test_roi_with_buy_sell_and_deposit():
    the_prices = {
        "XTZ": EUR(price=2),
        "ADA": EUR(price=3),

    }

    ada_trans_1 = user_transaction(user_id=1, volume=30, symbol="ADA", value=50,
                                   price=2, date="2020-07-15",
                                   source="kraken",
                                   source_id="",
                                   type="TRADE",
                                   order_type="BUY")

    ada_trans_2 = user_transaction(user_id=1, volume=20, symbol="ADA", value=40,
                                   price=2, date="2020-07-15",
                                   source="kraken",
                                   source_id="",
                                   type="TRADE",
                                   order_type="SELL")

    ada_trans_3 = user_transaction(user_id=1, volume=10, symbol="ADA", value=40,
                                   price=2, date="2020-07-15",
                                   source="kraken",
                                   source_id="",
                                   type="TRADE",
                                   order_type="BUY")

    deposit_trans = user_transaction(user_id=1, volume=13, symbol="EUR", value=40,
                                     price=1000, date="2020-07-15",
                                     source="kraken",
                                     source_id="",
                                     type="DEPOSIT",
                                     order_type="NONE")

    bc = BalanceCalculator(transactions=[ada_trans_1, ada_trans_2, ada_trans_3, deposit_trans],
                           preferred_currency="EUR", symbol_rates=the_prices, exchange_rates=[],
                           upper_bound_symbol_rates_date="2040-07-15", upper_bound_transaction_date="2040-07-15")

    out = bc.compute(user_id=1, date="2040-07-15")

    assert (len(out.user_grouped_symbol_values) == 1)

    assert out.user_grouped_symbol_values['ADA'].volume == 20

    assert out.user_grouped_symbol_values['ADA'].roi == (20 * 3 - (50 - 40 + 40)) / (50 - 40 + 40)

    assert out.roi == (20 * 3 - (50 - 40 + 40)) / (50 - 40 + 40)


def test_price_date_type():
    do_connect(configure_app())
    dt_now =int(datetime(year = 2021, month=1, day=1 ).timestamp()) * 1000
    prices.objects.all().delete()
    insert_prices_record_with_method(get_prices_with_int_date_record)
    prc = prices.objects(Q( status__timestamp__lt = dt_now)).order_by(
        '-status__timestamp')[:10]
    assert (len(prc) == 1)

    dt_now = int(datetime(year=2018, month=1, day=1).timestamp()) * 1000 #?
    prc = prices.objects(Q( status__timestamp__lt = dt_now)).order_by(
        '-status__timestamp')[:10]
    assert (len(prc) == 0)


from enum import Enum


class NOTIFICATION_TYPE(Enum):
    BALANCE = "BALANCE"
    SYMBOL_VALUE_DROP = "SYMBOL_VALUE_DROP"
    SYMBOL_VALUE_INCREASE = "SYMBOL_VALUE_INCREASE"

    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


import jsonpickle
from mongoengine import *

from calculator.helpers import if_none_raise
from cryptomodel.cryptostore import user_notification
from cryptomodel.notification_type import NOTIFICATION_TYPE

'''
Represents a a computed_user_notification 
Holds a notification 
Extends by holding the computed_date and the string value of json 
'''


class computed_notification(user_notification):
    meta = {'strict': False}
    computed_date = DateField()
    result = StringField()

    def set_result(self, value, computed_date):
        self.result = value
        self.computed_date = computed_date

    '''
    returns the equivalent object of the json stored in the result field  
    '''

    def get_result(self):
        if not hasattr(self, "result"):
            raise ValueError("does not exist : result ")
        if_none_raise(self.result)

        if self.notification_type == NOTIFICATION_TYPE.BALANCE.name:
            return jsonpickle.encode(self.result)
        else:
            raise NotImplementedError(self.notification_type)



from datetime import datetime

import jsonpickle

from cryptomodel.computed_notification import computed_notification
from cryptomodel.event_store import event_store, EVENT_TYPES, event_data
from cryptomodel.operations import OPERATIONS

DATE_FORMAT = '%Y-%m-%d'


def test_dynamic():
    cn = computed_notification()
    dt = datetime.now()
    cn.computed_date = dt
    es = event_store()
    es.key = 1
    ed = event_data()
    ed.type = EVENT_TYPES.COMPUTED_NOTIFICATION
    ed.data = jsonpickle.encode(cn)
    es.event_data = ed

    out = es.get_data_str_to_obj()
    assert (out.computed_date.year == dt.year)
    assert (out.computed_date.month == dt.month)
    assert (out.computed_date.day == dt.day)


def test_create():
    cn = computed_notification()
    cn.user_id = 1
    cn.computed_date = datetime.now()
    cn.result = "{a:1}"
    cn.check_every = "00:00"
    cn.is_active = False

    ed = event_data()
    ed.type = EVENT_TYPES.COMPUTED_NOTIFICATION
    ed.data =cn

    et = event_store.create(key=1, event_type=EVENT_TYPES.USER_NOTIFICATION, data=cn,operation=OPERATIONS.ADDED )
    cn2 = et.get_data_str_to_obj()

    assert (cn.computed_date.year == cn2.computed_date.year)
    assert (cn.check_every == cn2.check_every)





from datetime import datetime

import jsonpickle

from build.lib.cryptomodel.cryptostore_symbol_value import user_grouped_symbol_value, user_symbol_value
from cryptomodel.computed_notification import computed_notification
from cryptomodel.cryptostore import user_notification, user_transaction
from cryptomodel.cryptostore_symbol_value import total_symbol_value

DATE_FORMAT = '%Y-%m-%d'


def test_dynamic():

    cn = computed_notification()
    cn.notification_type = "BALANCE"
    cn.set_result(value=create_balance(), computed_date=datetime.now())
    out = cn.get_result()
    out2 = jsonpickle.decode(out)
    assert (out2.user_id == 1)


def create_balance():
    tsv = total_symbol_value(user_id=1, date_time_calculated=datetime.now(),
                             preferred_currency="EUR", symbol_prices=[], exchange_rates=[],
                             upper_bound_transaction_date=datetime.now(), upper_bound_symbol_rates_date=datetime.now()
                             )
    tsv.user_id = 1
    tsv.converted_value = 1
    tsv.transaction_value = 1

    t = user_transaction()
    t.user_id = 1
    t.value = 1
    t.volume = 1
    t.source = "kraken"
    t.symbol = "OXT"
    t.date = datetime.now()
    t.is_valid = True

    usv = user_symbol_value(user_id=1, converted_currency="EUR", present_price=1, transaction=t,
                            date_time_calculated=datetime.now(), exchange_rates=[], currency="EUR")

    usgv = user_grouped_symbol_value(user_id=1, symbol="OXT", volume=1, price=2, currency="EUR",
                                     date_time_calculated=datetime.now(),
                                     value_bought=1, converted_value=1)

    usgv.user_symbol_values.append(usv)

    usgv.value_bought = 1
    usgv.roi = .10

    tsv.user_grouped_symbol_values['OXT'] = usgv
    return tsv



import jsonpickle
from mongoengine import *

from calculator.helpers import if_none_raise
from cryptomodel.computed_notification import computed_notification
from cryptomodel.notification_type import NOTIFICATION_TYPE

'''
Represents a sent notification  
Holds a computed_notifications 
Extends with sent flag and sent_date
'''


class sent_notification(computed_notification):
    meta = {'strict': False}
    sent_date = DateField()



from calculator.CurrencyConverter import CurrencyConverter
from cryptomodel.order_types import ORDER_TYPES
from cryptomodel.transaction_types import TRANSACTION_TYPES

"""
A structure that holds all transactions per symbol 
computes the total value based on exchange and symbol rates 
contains a dict of user_grouped_symbol_value
user_grouped_symbol_value contains a list of user_symbol_value
also holds the passed in transactions that may be marked as invalid (symbol does not exist)
"""


class total_symbol_value(object):

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['exchange_rates']
        del state['symbol_prices']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __init__(self, user_id, date_time_calculated, preferred_currency, symbol_prices, exchange_rates,
                 upper_bound_symbol_rates_date, upper_bound_transaction_date):
        self.symbol_prices = symbol_prices
        self.exchange_rates = exchange_rates
        self.user_id = user_id
        self.converted_currency = preferred_currency
        self.converted_value = 0
        self.date_time_calculated = date_time_calculated
        self.upper_bound_symbol_rates_date = upper_bound_symbol_rates_date
        self.upper_bound_transaction_date = upper_bound_transaction_date
        self.user_grouped_symbol_values = {}
        self.transactions = []
        self.transaction_value = 0
        self.roi = 0

    def create(self, transactions):
        for trans in transactions:
            self.compute_transaction_value(trans)
        return self

    def compute_transaction_value(self, transaction):
        self.transactions.append(transaction)
        if not transaction.validate_symbol(self.symbol_prices):
            return

        if transaction.type == TRANSACTION_TYPES.TRADE.name:
            computed_usv = user_symbol_value(user_id=self.user_id, transaction=transaction,
                                             present_price=self.symbol_prices[transaction.symbol].price,
                                             currency=self.converted_currency, exchange_rates=self.exchange_rates,
                                             date_time_calculated=self.date_time_calculated,
                                             converted_currency=self.converted_currency)
            ugsv = self.contains_symbol(transaction)
            ugsv = self.create_user_group_symbol_value_or_increment(computed_usv, transaction, ugsv)

            self.converted_value += computed_usv.converted_value if transaction.order_type == ORDER_TYPES.BUY.name else -computed_usv.converted_value
            self.transaction_value += transaction.value if transaction.order_type == ORDER_TYPES.BUY.name else -transaction.value

            self.roi = (self.converted_value - self.transaction_value) / self.transaction_value
            ugsv.user_symbol_values.append(computed_usv)

    def create_user_group_symbol_value_or_increment(self, computed_usv, transaction, ugsv):

        if ugsv is not None:
            ugsv.volume += computed_usv.user_transaction.volume if transaction.order_type == ORDER_TYPES.BUY.name else -computed_usv.user_transaction.volume
            ugsv.converted_value += computed_usv.value if transaction.order_type == ORDER_TYPES.BUY.name else - computed_usv.value
            ugsv.converted_currency = self.converted_currency
            ugsv.value_bought += transaction.value if transaction.order_type == ORDER_TYPES.BUY.name else -transaction.value
            ugsv.roi = (ugsv.converted_value - ugsv.value_bought) / ugsv.value_bought

        else:
            ugsv = user_grouped_symbol_value(user_id=self.user_id,
                                             volume=transaction.volume,
                                             price=self.symbol_prices[transaction.symbol].price,
                                             currency=self.converted_currency,
                                             date_time_calculated=self.date_time_calculated,
                                             symbol=transaction.symbol,
                                             converted_value = computed_usv.value  if transaction.order_type == ORDER_TYPES.BUY.name else - computed_usv.value,
                                             value_bought=transaction.value if transaction.order_type == ORDER_TYPES.BUY.name else -transaction.value
                                             )
        self.user_grouped_symbol_values[transaction.symbol] = ugsv
        return ugsv

    def contains_symbol(self, transaction):
        if self.user_grouped_symbol_values is None:
            return None

        if not transaction.symbol in self.user_grouped_symbol_values.keys():
            return None

        return self.user_grouped_symbol_values[transaction.symbol]


class user_grouped_symbol_value:
    def __init__(self, user_id, symbol, volume, price, currency, date_time_calculated, value_bought, converted_value):
        self.symbol = symbol
        self.volume = volume
        self.date_time_calculated = date_time_calculated
        self.user_id = user_id
        self.currency = currency
        self.price = price
        self.converted_currency = ""
        self.converted_value = converted_value
        self.user_symbol_values = []
        self.value_bought = value_bought
        self.roi = (self.converted_value - self.value_bought) / self.value_bought


class user_symbol_value:
    def __init__(self, user_id, transaction, present_price,
                 currency, date_time_calculated, converted_currency, exchange_rates):
        self.user_transaction = transaction
        self.present_price = present_price
        self.currency = currency
        self.date_time_calculated = date_time_calculated
        self.value = self.present_price * self.user_transaction.volume
        self.user_id = user_id
        self.converted_currency = converted_currency
        self.exchange_rates = exchange_rates
        cc = CurrencyConverter(exchange_rates)
        self.converted_value = cc.convert_value(self.value, self.currency, self.converted_currency)
        self.roi = (self.value - self.user_transaction.value) / self.user_transaction.value



from keyring import get_password
from mongoengine import *

from cryptomodel.channel_type import CHANNEL_TYPE
from cryptomodel.notification_type import NOTIFICATION_TYPE
from cryptomodel.operations import OPERATIONS
from cryptomodel.order_types import ORDER_TYPES
from cryptomodel.transaction_types import TRANSACTION_TYPES


class user_settings(Document):
    meta = {'strict': False}
    user_id = IntField()
    preferred_currency = StringField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())


''' Represents a user notification
    notification_type:
        balance : computes the balance
        symbol drop : sends all symbols that drop with value less than threshold_value
        symbol increase : sends all symbols that increase with value more than threshold_value
    
    channel type keeps the channel used for notifying 
        only telegram  is supported 
        
    user_id to check for 

    start_date : date time to first check 
    end_date : last date time to check 
    check_every : will check some time and then every hh:mm after (00:00 for no re-check)
    
    threshold_value : some notifications check against some value (see symbol drop or increase )
    is_active : if false will not check at all

    source_id : the original _id of the record copied here (used when one service messages another  
    (destination needs to know the original source id )
    
'''


class user_notification(Document):
    meta = {'strict': False, 'allow_inheritance': True}
    user_id = IntField()
    user_name = StringField()
    user_email = StringField()
    notification_type = StringField(choices=NOTIFICATION_TYPE.choices())
    check_every = StringField()
    start_date = DateField()
    end_date = DateField()
    is_active = BooleanField()
    channel_type = StringField(choices=CHANNEL_TYPE.choices())
    threshold_value = LongField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())


class user_transaction(Document):
    meta = {'strict': False}
    user_id = IntField()
    volume = FloatField()
    symbol = StringField()
    value = FloatField()
    price = FloatField()
    date = DateField()
    source = StringField()
    currency = StringField()
    order_type = StringField(choices=ORDER_TYPES.choices())
    type = StringField(choices=TRANSACTION_TYPES.choices())
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())
    is_valid = BooleanField(default=True)
    invalid_reason = StringField(default="")

    def validate_symbol(self, symbols):
        if symbols is None:
            self.is_valid = False
            self.invalid_reason = "symbol list empty " + self.symbol
            return self.is_valid

        if self.symbol not in symbols.keys():
            self.is_valid = False
            self.invalid_reason = "symbol does not exist in list value:" + self.symbol
        return self.is_valid


class user_channel(Document):
    meta = {'strict': False}
    user_id = IntField()
    channel_type = StringField()
    chat_id = StringField()
    user_email = StringField()
    source_id = ObjectIdField()
    operation = StringField(choices=OPERATIONS.choices())

    def set_token(self):
        self.token = get_password(self.notification_type, 'token')



import jsonpickle
from mongoengine import *

from enum import Enum

from calculator.helpers import if_empty_string_raise
from cryptomodel.operations import OPERATIONS


class EVENT_TYPES(Enum):
    USER_NOTIFICATION = "USER_NOTIFICATION"
    COMPUTED_NOTIFICATION = "COMPUTED_NOTIFICATION"
    SENT_NOTIFICATION = "SENT_NOTIFICATION"
    TRANSACTIONS = "TRANSACTIONS"
    SETTINGS = "SETTINGS"


    @classmethod
    def choices(cls):
        return tuple((i.name, i.value) for i in cls)


class event_data(EmbeddedDocument):
    type = StringField(choices=EVENT_TYPES.choices())
    operation = StringField(operations=OPERATIONS.choices())
    data = StringField()

'''
Represents an event item for event based communication between microservices 
where key is user_id (used for petitioning)
and data is the string equivalent of the json representation of a data and the event_type 
'''


class event_store(Document):
    meta = {'strict': False}
    key = StringField()
    event_data = EmbeddedDocumentField(event_data, db_field='event_data')

    '''decodes the string json of data to object '''
    def get_data_str_to_obj(self):
        if_empty_string_raise(self.event_data)
        if_empty_string_raise(self.event_data.data)
        return jsonpickle.decode(self.event_data.data)

    @staticmethod
    def create(key, event_type, data, operation):
        es  = event_store()
        es.key = key
        ed = event_data()
        ed.operation = operation
        ed.data = jsonpickle.encode(data)
        ed.type = event_type
        es.event_data = ed
        return es